// XKList.c - List widget

#include <stdio.h>
#include <ctype.h>

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>

#include <X11/Xmu/Drawing.h>

#include <X11/neXtaw/XawInit.h>
#include <XKListP.h>

// These added so widget knows whether its height, width are user selected.
// I also added the freedoms member of the list widget part.
// I assume this was Alfredo?

#define HeightLock  1
#define WidthLock   2
#define LongestLock 4

#define HeightFree(w)  !(((XKListWidget)(w))->xklist.freedoms & HeightLock)
#define WidthFree(w)   !(((XKListWidget)(w))->xklist.freedoms & WidthLock)
#define LongestFree(w) !(((XKListWidget)(w))->xklist.freedoms & LongestLock)

// Default Translation table.

static char defaultTranslations[] =  
  "<Btn1Down>:   Set()\n\
   <Btn1Up>:     Notify()";

// Private Data

#define offset(field) XtOffset(XKListWidget, field)

static XtResource resources[] =
{
    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
		offset(xklist.foreground), XtRString, XtDefaultForeground },
    { XtNcursor, XtCCursor, XtRCursor, sizeof(Cursor),
       offset(simple.cursor), XtRString, "left_ptr" },
    { XtNfont,  XtCFont, XtRFontStruct, sizeof(XFontStruct *),
		offset(xklist.font),XtRString, XtDefaultFont },
    { XtNfontSet,  XtCFontSet, XtRFontSet, sizeof(XFontSet ),
		offset(xklist.fontset),XtRString, XtDefaultFontSet },
    { XtNlist, XtCList, XtRPointer, sizeof(char **),
       offset(xklist.list), XtRString, NULL },
    { XtNdefaultColumns, XtCColumns, XtRInt,  sizeof(int),
		offset(xklist.default_cols), XtRImmediate, (XtPointer)2 },
    { XtNlongest, XtCLongest, XtRInt,  sizeof(int),
		offset(xklist.longest), XtRImmediate, (XtPointer)0 },
    { XtNnumberStrings, XtCNumberStrings, XtRInt,  sizeof(int),
		offset(xklist.nitems), XtRImmediate, (XtPointer)0 },
    { XtNpasteBuffer, XtCBoolean, XtRBoolean,  sizeof(Boolean),
		offset(xklist.paste), XtRImmediate, (XtPointer) False },
    { XtNforceColumns, XtCColumns, XtRBoolean,  sizeof(Boolean),
		offset(xklist.force_cols), XtRImmediate, (XtPointer) False },
    { XtNverticalList, XtCBoolean, XtRBoolean,  sizeof(Boolean),
		offset(xklist.vertical_cols), XtRImmediate, (XtPointer) False },
    { XtNinternalWidth, XtCWidth, XtRDimension,  sizeof(Dimension),
		offset(xklist.internal_width), XtRImmediate, (XtPointer) 2 },
    { XtNinternalHeight, XtCHeight, XtRDimension, sizeof(Dimension),
		offset(xklist.internal_height), XtRImmediate, (XtPointer) 2 },
    { XtNcolumnSpacing, XtCSpacing, XtRDimension,  sizeof(Dimension),
		offset(xklist.column_space), XtRImmediate, (XtPointer) 6 },
    { XtNrowSpacing, XtCSpacing, XtRDimension,  sizeof(Dimension),
		offset(xklist.row_space), XtRImmediate, (XtPointer) 2 },
    { XtNcallback, XtCCallback, XtRCallback, sizeof(XtPointer),
        offset(xklist.callback), XtRCallback, NULL },
};

static	void				Initialize();
static	void				ChangeSize(Widget, Dimension, Dimension);
static	void				Resize();
static	void				Redisplay();
static	void				Destroy();
static	Boolean				Layout();
static	XtGeometryResult	PreferredGeom();
static	Boolean				SetValues();
static	void				Notify();
static	void				Set();
static	void				Unset();

static	XtActionsRec	actions[] =
{
      { "Notify",         Notify },
      { "Set",            Set },
      { "Unset",          Unset },
};

XKListClassRec xkListClassRec =
{
	{ // core_class fields
		/* superclass	  			*/	(WidgetClass) &simpleClassRec,
		/* class_name	  			*/	"XKList",
		/* widget_size	  			*/	sizeof(XKListRec),
		/* class_initialize  		*/	XawInitializeWidgetSet,
		/* class_part_initialize	*/	NULL,
		/* class_inited       		*/	FALSE,
		/* initialize	  			*/	Initialize,
		/* initialize_hook			*/	NULL,
		/* realize		  			*/	XtInheritRealize,
		/* actions		  			*/	actions,
		/* num_actions	  			*/	XtNumber(actions),
		/* resources	  			*/	resources,
		/* num_resources	  		*/	XtNumber(resources),
		/* xrm_class	  			*/	NULLQUARK,
		/* compress_motion	  		*/	TRUE,
		/* compress_exposure  		*/	FALSE,
		/* compress_enterleave		*/	TRUE,
		/* visible_interest	  		*/	FALSE,
		/* destroy		  			*/	Destroy,
		/* resize		  			*/	Resize,
		/* expose		  			*/	Redisplay,
		/* set_values	  			*/	SetValues,
		/* set_values_hook			*/	NULL,
		/* set_values_almost		*/	XtInheritSetValuesAlmost,
		/* get_values_hook			*/	NULL,
		/* accept_focus	 			*/	NULL,
		/* version					*/	XtVersion,
		/* callback_private   		*/	NULL,
		/* tm_table		   			*/	defaultTranslations,
		/* query_geometry			*/  PreferredGeom,
	},
	{ // Simple class fields initialization
		/* change_sensitive			*/	XtInheritChangeSensitive
	},
	{ // XKList class fields initialization
		/* not used					*/	0
	},
};

WidgetClass xkListWidgetClass = (WidgetClass) &xkListClassRec;

static void GetGCs(w)
Widget w;
{
	Display			*display = XtDisplay(w);
	int				screen = DefaultScreen(display);
    XGCValues		values;
    XKListWidget	lw = (XKListWidget) w;
	XColor			grey35, xexact;

    values.foreground = lw->xklist.foreground;
    values.font	= lw->xklist.font->fid;
    if ((lw->simple.international == True))
	{
        lw->xklist.normgc = XtAllocateGC(w, 0, (unsigned) GCForeground,
			 &values, GCFont, 0);
	}
    else
		{ lw->xklist.normgc = XtGetGC(w, (GCForeground | GCFont), &values); }

    values.foreground = lw->core.background_pixel;
    if ((lw->simple.international == True))
	{
        lw->xklist.revgc = XtAllocateGC(w, 0, (unsigned) GCForeground,
			 &values, GCFont, 0);
	}
    else
		{ lw->xklist.revgc = XtGetGC(w, (GCForeground | GCFont), &values); }

    values.foreground = WhitePixel(display, screen);
    if ((lw->simple.international == True))
	{
        lw->xklist.whitegc = XtAllocateGC(w, 0, (unsigned) GCForeground,
			 &values, GCFont, 0);
	}
    else
		{ lw->xklist.whitegc = XtGetGC(w, (GCForeground | GCFont), &values); }

	values.tile = XmuCreateStippledPixmap(XtScreen(w), 
		lw->xklist.foreground,
		lw->core.background_pixel,
		lw->core.depth);
    values.fill_style = FillTiled;

    if ((lw->simple.international == True))
	{
        lw->xklist.graygc = XtAllocateGC(w, 0, (unsigned) GCTile | GCFillStyle,
			&values, GCFont, 0);
	}
    else
	{
        lw->xklist.graygc = XtGetGC(w, (unsigned) GCFont | GCTile | GCFillStyle,
			  &values);
	}

	XAllocNamedColor(display,
		DefaultColormapOfScreen(DefaultScreenOfDisplay(display)),
		"grey35",
		&grey35, &xexact);
	values.foreground = grey35.pixel;
	lw->xklist.dimgc = XtGetGC(w, (GCForeground), &values);
}

// Does routine checks/computations that must be done after data changes
// but won't hurt if accidently called. These calculations were needed in
// SetValues. They were in ResetList. ResetList called ChangeSize, which 
// made an XtGeometryRequest. You MAY NOT change your geometry from within
// a SetValues. (Xt man, sect. 9.7.2)  So, I factored these changes out.
static void CalculatedValues(Widget w)
{
    XKListWidget	lw = (XKListWidget) w;
	Widget			parent = XtParent(w);
    int				i, len;

    // If xklist is NULL then the xklist will just be the name of the widget.
    if (!(lw->xklist.list))
		{ lw->xklist.list = &(lw->core.name); lw->xklist.nitems = 1; }

    // Get number of items.
    if (!(lw->xklist.nitems))
	{
        for ( ;
			lw->xklist.list[lw->xklist.nitems];
			lw->xklist.nitems++);
	}

    // Get column width.
    if (LongestFree(lw))
	{
        lw->xklist.longest = 0; // So it will accumulate real longest below.

        for (i = 0; i < lw->xklist.nitems; i++)
		{
            if ((lw->simple.international == True))
			{
				len = XmbTextEscapement(lw->xklist.fontset, lw->xklist.list[i],
					strlen(lw->xklist.list[i]));
			}
            else
			{
                len = XTextWidth(lw->xklist.font, lw->xklist.list[i],
					strlen(lw->xklist.list[i]));
			}
            if (len > lw->xklist.longest) { lw->xklist.longest = len; }
        }
    }

	// Offset of 12 for arrows. Try to stretch this out for use as a single
	// column list: e.g. a directory listing. At the moment this subtracts 
	// 20 for the scrollbar of a viewport: for some reason getting the parent
	// width doesn't seem to get the clip width.
	if ((parent) && XtIsRealized(parent) &&
		(lw->xklist.force_cols == True) &&
		(lw->xklist.default_cols == 1))
	{
		Dimension	width;

		XtVaGetValues(parent, XtNwidth, (XtArgVal) &width, NULL);
		lw->xklist.col_width = 
			(lw->xklist.longest + lw->xklist.column_space + 12 > width - 32)
					? lw->xklist.longest + lw->xklist.column_space + 12
					: width - 32;
	}
	else
	{
		lw->xklist.col_width =
			lw->xklist.longest + lw->xklist.column_space + 12;
	}
}

// Resets the new xklist when important things change. Arguments: w - the
// widget. changex, changey - allow the height or width to change?
// Returns: TRUE if width or height have been changed.
static void ResetList(Widget w, Boolean changex, Boolean changey)
{
    Dimension	width = w->core.width;
    Dimension	height = w->core.height;

    CalculatedValues(w);
    if (Layout(w, changex, changey, &width, &height))
		{ ChangeSize(w, width, height); }
}

// Layout the widget.
static void ChangeSize(Widget w, Dimension width, Dimension height)
{
    XtWidgetGeometry request, reply;

    request.request_mode = CWWidth | CWHeight;
    request.width = width;
    request.height = height;

    switch (XtMakeGeometryRequest(w, &request, &reply))
	{
		case XtGeometryYes:
			break;
		case XtGeometryNo:
			break;
		case XtGeometryAlmost:
			Layout(w, (request.height != reply.height),
				(request.width != reply.width),
				&(reply.width), &(reply.height));
			request = reply;
			switch (XtMakeGeometryRequest(w, &request, &reply))
			{
				case XtGeometryYes:
				case XtGeometryNo:
					break;
				case XtGeometryAlmost:
					request = reply;
					Layout(w, FALSE, FALSE,
						&(request.width), &(request.height));
					request.request_mode = CWWidth | CWHeight;
					XtMakeGeometryRequest(w, &request, &reply);
					break;
				default:
					XtAppWarning(XtWidgetToApplicationContext(w),
						"List Widget: Unknown geometry return.");
					break;
			}
			break;
		default:
			XtAppWarning(XtWidgetToApplicationContext(w),
				"List Widget: Unknown geometry return.");
			break;
    }
}

/*	Function Name: Initialize
 *	Description: Function that initilizes the widget instance.
 *	Arguments: junk - NOT USED.
 *                 new  - the new widget.
 *	Returns: none
 */

/* ARGSUSED */
static void 
Initialize(junk, new, args, num_args)
Widget junk, new;
ArgList args;
Cardinal *num_args;
{
    XKListWidget lw = (XKListWidget) new;

	// Initialize all private resources: record for posterity if we are free.
    lw->xklist.freedoms = (lw->core.width != 0) * WidthLock +
		(lw->core.height != 0) * HeightLock +
		(lw->xklist.longest != 0) * LongestLock;

    GetGCs(new);

    // Set row height. based on font or fontset.
    if ((lw->simple.international == True))
	{
        lw->xklist.row_height =
			XExtentsOfFontSet(lw->xklist.fontset)->max_ink_extent.height
				+ lw->xklist.row_space;
	}
    else
	{
        lw->xklist.row_height = lw->xklist.font->max_bounds.ascent
			+ lw->xklist.font->max_bounds.descent
			+ lw->xklist.row_space;
	}
    ResetList(new, WidthFree(lw), HeightFree(lw));
    lw->xklist.highlight = lw->xklist.is_highlighted = NO_HIGHLIGHT;

	lw->xklist.arrows = NULL;
}

/*	Function Name: CvtToItem
 *	Description: Converts Xcoord to item number of item containing that
 *                   point.
 *	Arguments: w - the xklist widget.
 *                 xloc, yloc - x location, and y location.
 *	Returns: the item number.
 */

static int
CvtToItem(w, xloc, yloc, item)
Widget w;
int xloc, yloc;
int *item;
{
    int one, another;
    XKListWidget lw = (XKListWidget) w;
    int ret_val = OKAY;

    if (lw->xklist.vertical_cols)
	{
        one = lw->xklist.nrows * ((xloc - (int) lw->xklist.internal_width)
			/ lw->xklist.col_width);
        another = (yloc - (int) lw->xklist.internal_height) 
	        / lw->xklist.row_height;
		// If out of range, return minimum possible value.
		if (another >= lw->xklist.nrows)
			{ another = lw->xklist.nrows - 1; ret_val = OUT_OF_RANGE; }
    }
    else
	{
        one = (lw->xklist.ncols * ((yloc - (int) lw->xklist.internal_height) 
			/ lw->xklist.row_height)) ;
		// If in right margin handle things right.
        another = (xloc - (int) lw->xklist.internal_width) /
			lw->xklist.col_width;
		if (another >= lw->xklist.ncols)
			{ another = lw->xklist.ncols - 1; ret_val = OUT_OF_RANGE; }
    }  
    if ((xloc < 0) || (yloc < 0)) { ret_val = OUT_OF_RANGE; }
    if (one < 0) { one = 0; }
    if (another < 0) { another = 0; }
    *item = one + another;
    if (*item >= lw->xklist.nitems) { return OUT_OF_RANGE; }
    return ret_val;
}

/*	Function Name: FindCornerItems.
 *	Description: Find the corners of the rectangle in item space.
 *	Arguments: w - the xklist widget.
 *                 event - the event structure that has the rectangle it it.
 *                 ul_ret, lr_ret - the corners ** RETURNED **.
 *	Returns: none.
 */

static void
FindCornerItems(w, event, ul_ret, lr_ret)
Widget w;
XEvent * event;
int *ul_ret, *lr_ret;
{
    int xloc, yloc;

    xloc = event->xexpose.x;
    yloc = event->xexpose.y;
    CvtToItem(w, xloc, yloc, ul_ret);
    xloc += event->xexpose.width;
    yloc += event->xexpose.height;
    CvtToItem(w, xloc, yloc, lr_ret);
}

/*	Function Name: ItemInRectangle
 *	Description: returns TRUE if the item passed is in the given rectangle.
 *	Arguments: w - the xklist widget.
 *                 ul, lr - corners of the rectangle in item space.
 *                 item - item to check.
 *	Returns: TRUE if the item passed is in the given rectangle.
 */

static Boolean
ItemInRectangle(w, ul, lr, item)
Widget w;
int ul, lr, item;
{
    XKListWidget	lw = (XKListWidget) w;
    int				mod_item;
    int				things;
    
    if (item < ul || item > lr) 
        { return FALSE; }
    if (lw->xklist.vertical_cols)
        { things = lw->xklist.nrows; }
    else
        { things = lw->xklist.ncols; }

    mod_item = item % things;
    if ((mod_item >= ul % things) && (mod_item <= lr % things))
        { return TRUE; }
    return FALSE;
}

// Paints the color of the background for the given item. It performs
// clipping to the interior of internal_width/height by hand, as its a
// simple calculation and probably much faster than using Xlib and a clip mask.
//
//  x, y - ul corner of the area item occupies.
//  gc - the gc to use to paint this rectangle */

static void
HighlightBackground(w, x, y, gc)
Widget w;
int x, y;
GC gc;
{
    XKListWidget lw = (XKListWidget) w;

    // Easy to clip the rectangle by hand and probably alot faster than Xlib.
    Dimension width               = lw->xklist.col_width;
    Dimension height              = lw->xklist.row_height;
    Dimension frame_limited_width = w->core.width -
									lw->xklist.internal_width -
									x;
    Dimension frame_limited_height= w->core.height -
									lw->xklist.internal_height -
									y;

    // Clip the rectangle width and height to the edge of the drawable area.

    if  (width > frame_limited_width)
        { width = frame_limited_width; }
    if  (height > frame_limited_height)
        { height = frame_limited_height; }

    // Clip the rectangle x and y to the edge of the drawable area.
    if (x < lw->xklist.internal_width)
	{
        width = width - (lw->xklist.internal_width - x);
        x = lw->xklist.internal_width;
    }
    if (y < lw->xklist.internal_height)
	{
        height = height - (lw->xklist.internal_height - x);
        y = lw->xklist.internal_height;
    }
    XFillRectangle(XtDisplay(w), XtWindow(w), gc, x, y, width, height);
}


/* ClipToShadowInteriorAndLongest()
 *
 * Converts the passed gc so that any drawing done with that GC will not
 * write in the empty margin (specified by internal_width/height) (which also
 * prevents erasing the shadow.  It also clips against the value longest.
 * If the user doesn't set longest, this has no effect (as longest is the
 * maximum of all item lengths).  If the user does specify, say, 80 pixel
 * columns, though, this prevents items from overwriting other items. */

static void ClipToShadowInteriorAndLongest(lw, gc_p, x)
    XKListWidget lw; 
    GC* gc_p;
    Dimension x;
{
    XRectangle rect;

    rect.x = x;
    rect.y = lw->xklist.internal_height;
    rect.height = lw->core.height - lw->xklist.internal_height * 2;
    rect.width = lw->core.width - lw->xklist.internal_width - x;
	// #### try to remove longest clipping maybe should be columnwidth?
    // if (rect.width > lw->xklist.longest)
    //     { rect.width = lw->xklist.longest; }

    XSetClipRectangles(XtDisplay((Widget) lw), *gc_p, 0, 0, &rect, 1, YXBanded);
}


// Paints the name of the item in the appropriate location.
static void PaintItemName(Widget w, int item)
{
	Display			*display = XtDisplay(w);
	Window			window;
    XKListWidget	lw = (XKListWidget) w;
    XFontSetExtents	*ext  = XExtentsOfFontSet(lw->xklist.fontset);
    char			*str;
    GC				gc;
    int				x, y, str_y;

    if (!XtIsRealized(w)) { return; }	 // Just in case ...

	window = XtWindow(w);

    if ((lw->xklist.vertical_cols))
	{
		x = lw->xklist.col_width * (item / lw->xklist.nrows)
			+ lw->xklist.internal_width;
		y = lw->xklist.row_height * (item % lw->xklist.nrows)
			+ lw->xklist.internal_height;
    }
    else
	{
        x = lw->xklist.col_width * (item % lw->xklist.ncols)
			+ lw->xklist.internal_width;
        y = lw->xklist.row_height * (item / lw->xklist.ncols)
			+ lw->xklist.internal_height;
    }

    if ((lw->simple.international == True))
        { str_y = y + abs(ext->max_ink_extent.y); }
    else
        { str_y = y + lw->xklist.font->max_bounds.ascent; }

    if ((item == lw->xklist.is_highlighted))
	{
        if ((item == lw->xklist.highlight))
		{
			gc = lw->xklist.normgc;
            // gc = lw->xklist.revgc;
			// HighlightBackground(w, x, y, lw->xklist.normgc);
			HighlightBackground(w, x, y, lw->xklist.whitegc);
		}
        else
		{
			if (XtIsSensitive(w)) 
				{ gc = lw->xklist.normgc; }
			else
				{ gc = lw->xklist.graygc; }
			HighlightBackground(w, x, y, lw->xklist.revgc);
			lw->xklist.is_highlighted = NO_HIGHLIGHT;
        }
    }
    else
	{
		if (item == lw->xklist.highlight)
		{
			gc = lw->xklist.normgc;
			// #### HighlightBackground(w, x, y, lw->xklist.normgc);
			HighlightBackground(w, x, y, lw->xklist.whitegc);
			lw->xklist.is_highlighted = item;
		}
		else
		{
			if (XtIsSensitive(w)) 
				{ gc = lw->xklist.normgc; }
			else
				{ gc = lw->xklist.graygc; }
		}
    }

    // List's overall width contains the same number of inter-column
    // column_space's as columns. There should thus be a half column_width
	// margin on each side of each column. The row case is symmetric.

    x += lw->xklist.column_space / 2;
    str_y += lw->xklist.row_space / 2;
    str = lw->xklist.list[item];			// draw it

    ClipToShadowInteriorAndLongest(lw, &gc, x);

    if ((lw->simple.international == True))
	{
        XmbDrawString(display, window, lw->xklist.fontset,
			gc, x, str_y, str, strlen(str));
	}
    else
	{
        XDrawString(display, window, gc, x, str_y, str, strlen(str));
	}

	// Only one column at the moment so ...
    if ((lw->xklist.arrows) && (lw->xklist.arrows[item] == True))
	{
		Position	w = lw->xklist.col_width - lw->xklist.column_space / 2;
		Position	y = str_y - lw->xklist.row_space / 2 - 6;
		Dimension	h = 6;

		// Draw the cascade indicator.
		XDrawLine(display, window, lw->xklist.dimgc,
			w - 5, y + 1, w, y + h / 2 - 1);
		XDrawLine(display, window, lw->xklist.whitegc,
			w - 6, y + h, w, y + h / 2);
		XDrawLine(display, window, lw->xklist.normgc,
			w - 6, y, w - 6, y + h);
    }
    XSetClipMask(display, gc, None);
}

// Repaints the widget window on expose events.
/* ARGSUSED */
static void 
Redisplay(w, event, junk)
Widget w;
XEvent *event;
Region junk;
{
    XKListWidget	lw = (XKListWidget) w;
    int				item;				// An item to work with.
    int				ul_item, lr_item;	// Corners of items we need to paint.

    if (!(event))						// repaint all.
	{
        ul_item = 0;
		lr_item = lw->xklist.nrows * lw->xklist.ncols - 1;
		XClearWindow(XtDisplay(w), XtWindow(w));
    }
    else
	{
        FindCornerItems(w, event, &ul_item, &lr_item);
 	}

    for (item = ul_item;
		(item <= lr_item && item < lw->xklist.nitems);
		++item)
	{
		if (ItemInRectangle(w, ul_item, lr_item, item))
			{ PaintItemName(w, item); }
	}
}


/* PreferredGeom()
 *
 * This tells the parent what size we would like to be
 * given certain constraints.
 * w - the widget.
 * intended - what the parent intends to do with us.
 * requested - what we want to happen. */

static XtGeometryResult 
PreferredGeom(w, intended, requested)
Widget w;
XtWidgetGeometry *intended, *requested;
{
    Dimension new_width, new_height;
    Boolean change, width_req, height_req;
    
    width_req = intended->request_mode & CWWidth;
    height_req = intended->request_mode & CWHeight;

    if (width_req)
      new_width = intended->width;
    else
      new_width = w->core.width;

    if (height_req)
      new_height = intended->height;
    else
      new_height = w->core.height;

    requested->request_mode = 0;
    
/*
 * We only care about our height and width.
 */

    if ( !width_req && !height_req)
      return(XtGeometryYes);
    
    change = Layout(w, !width_req, !height_req, &new_width, &new_height);

    requested->request_mode |= CWWidth;
    requested->width = new_width;
    requested->request_mode |= CWHeight;
    requested->height = new_height;

    if (change)
        return(XtGeometryAlmost);
    return(XtGeometryYes);
}


/* Resize()
 *
 * resizes the widget, by changing the number of rows and columns. */

static void
Resize(w)
    Widget w;
{
    Dimension width, height;

    width = w->core.width;
    height = w->core.height;

    if (Layout(w, FALSE, FALSE, &width, &height))
	XtAppWarning(XtWidgetToApplicationContext(w),
	   "List Widget: Size changed when it shouldn't have when resising.");
}


/* Layout()
 *
 * lays out the item in the xklist.
 * w - the widget.
 * xfree, yfree - TRUE if we are free to resize the widget in
 *                this direction.
 * width, height- the is the current width and height that we are going
 *                we are going to layout the xklist widget to,
 *                depending on xfree and yfree of course.
 *                               
 * RETURNS: TRUE if width or height have been changed. */

static Boolean
Layout(w, xfree, yfree, width, height)
Widget w;
Boolean xfree, yfree;
Dimension *width, *height;
{
    XKListWidget lw = (XKListWidget) w;
    Boolean change = FALSE;
    
/* 
 * If force columns is set then always use number of columns specified
 * by default_cols.
 */

    if (lw->xklist.force_cols) {
        lw->xklist.ncols = lw->xklist.default_cols;
	if (lw->xklist.ncols <= 0) lw->xklist.ncols = 1;
	/* 12/3 = 4 and 10/3 = 4, but 9/3 = 3 */
	lw->xklist.nrows = ( ( lw->xklist.nitems - 1) / lw->xklist.ncols) + 1 ;
	if (xfree) {		/* If allowed resize width. */

            /* this counts the same number
            of inter-column column_space 's as columns.  There should thus be a
            half column_space margin on each side of each column...*/

	    *width = lw->xklist.ncols * lw->xklist.col_width
	           + 2 * lw->xklist.internal_width;
	    change = TRUE;
	}
	if (yfree) {		/* If allowed resize height. */
	    *height = (lw->xklist.nrows * lw->xklist.row_height)
                    + 2 * lw->xklist.internal_height;
	    change = TRUE;
	}
	return(change);
    }

/*
 * If both width and height are free to change the use default_cols
 * to determine the number columns and set new width and height to
 * just fit the window.
 */

    if (xfree && yfree) {
        lw->xklist.ncols = lw->xklist.default_cols;
	if (lw->xklist.ncols <= 0) lw->xklist.ncols = 1;
	lw->xklist.nrows = ( ( lw->xklist.nitems - 1) / lw->xklist.ncols) + 1 ;
        *width = lw->xklist.ncols * lw->xklist.col_width
	       + 2 * lw->xklist.internal_width;
	*height = (lw->xklist.nrows * lw->xklist.row_height)
                + 2 * lw->xklist.internal_height;
	change = TRUE;
    }
/* 
 * If the width is fixed then use it to determine the number of columns.
 * If the height is free to move (width still fixed) then resize the height
 * of the widget to fit the current xklist exactly.
 */
    else if (!xfree) {
        lw->xklist.ncols = ( (int)(*width - 2 * lw->xklist.internal_width)
	                    / (int)lw->xklist.col_width);
	if (lw->xklist.ncols <= 0) lw->xklist.ncols = 1;
	lw->xklist.nrows = ( ( lw->xklist.nitems - 1) / lw->xklist.ncols) + 1 ;
	if ( yfree ) {
  	    *height = (lw->xklist.nrows * lw->xklist.row_height)
		    + 2 * lw->xklist.internal_height;
	    change = TRUE;
	}
    }
/* 
 * The last case is xfree and !yfree we use the height to determine
 * the number of rows and then set the width to just fit the resulting
 * number of columns.
 */
    else if (!yfree) {		/* xfree must be TRUE. */
        lw->xklist.nrows = (int)(*height - 2 * lw->xklist.internal_height) 
	                 / (int)lw->xklist.row_height;
	if (lw->xklist.nrows <= 0) lw->xklist.nrows = 1;
	lw->xklist.ncols = (( lw->xklist.nitems - 1 ) / lw->xklist.nrows) + 1;
	*width = lw->xklist.ncols * lw->xklist.col_width 
	       + 2 * lw->xklist.internal_width;
	change = TRUE;
    }      
    return(change);
}


/* Notify() - ACTION
 *
 * Notifies the user that a button has been pressed, and
 * calls the callback; if the XtNpasteBuffer resource is true
 * then the name of the item is also put in CUT_BUFFER0.	*/

/* ARGSUSED */
static void
Notify(w, event, params, num_params)
Widget w;
XEvent * event;
String * params;
Cardinal *num_params;
{
    XKListWidget lw = ( XKListWidget ) w;
    int item, item_len;
    XKListReturnStruct ret_value;

/* 
 * Find item and if out of range then unhighlight and return. 
 * 
 * If the current item is unhighlighted then the user has aborted the
 * notify, so unhighlight and return.
 */

    if ( ((CvtToItem(w, event->xbutton.x, event->xbutton.y, &item))
	  == OUT_OF_RANGE) || (lw->xklist.highlight != item) ) {
        XKListUnhighlight(w);
        return;
    }

    item_len = strlen(lw->xklist.list[item]);

    if ( lw->xklist.paste )	/* if XtNpasteBuffer set then paste it. */
        XStoreBytes(XtDisplay(w), lw->xklist.list[item], item_len);

/* 
 * Call Callback function.
 */

    ret_value.string = lw->xklist.list[item];
    ret_value.list_index = item;
    
    XtCallCallbacks( w, XtNcallback, (XtPointer) &ret_value);
}


// Unhighlights the current element.
/* ARGSUSED */
static void
Unset(w, event, params, num_params)
Widget w;
XEvent * event;
String * params;
Cardinal *num_params;
{
  XKListUnhighlight(w);
}


// Highlights the current element.
/* ARGSUSED */
static void
Set(w, event, params, num_params)
Widget w;
XEvent * event;
String * params;
Cardinal *num_params;
{
	XKListWidget	lw = (XKListWidget) w;
	int				item;

	if ((CvtToItem(w, event->xbutton.x, event->xbutton.y, &item))
		== OUT_OF_RANGE)
		{ XKListUnhighlight(w);	}	        // Unhighlight current item.
	else if (lw->xklist.is_highlighted != item)
    	{ XKListHighlight(w, item); }	    // If not highlighted then ...
}

// Set specified arguments into widget

static Boolean 
SetValues(current, request, new, args, num_args)
Widget current, request, new;
ArgList args;
Cardinal *num_args;
{
    XKListWidget cl = (XKListWidget) current;
    XKListWidget rl = (XKListWidget) request;
    XKListWidget nl = (XKListWidget) new;
    Boolean redraw = FALSE;
    XFontSetExtents *ext = XExtentsOfFontSet(nl->xklist.fontset);

    /* If the request height/width is different, lock it.  Unless its 0. If */
    /* neither new nor 0, leave it as it was.  Not in R5. */
    if ( nl->core.width != cl->core.width )
        nl->xklist.freedoms |= WidthLock;
    if ( nl->core.width == 0 )
        nl->xklist.freedoms &= ~WidthLock;

    if ( nl->core.height != cl->core.height )
        nl->xklist.freedoms |= HeightLock;
    if ( nl->core.height == 0 )
        nl->xklist.freedoms &= ~HeightLock;

    if ( nl->xklist.longest != cl->xklist.longest )
        nl->xklist.freedoms |= LongestLock;
    if ( nl->xklist.longest == 0 )
        nl->xklist.freedoms &= ~LongestLock;

    /* _DONT_ check for fontset here - it's not in GC.*/

    if ((cl->xklist.foreground != nl->xklist.foreground) ||
		(cl->core.background_pixel != nl->core.background_pixel) ||
		(cl->xklist.font != nl->xklist.font))
	{
		XGCValues values;
		XGetGCValues(XtDisplay(current), cl->xklist.graygc, GCTile, &values);
		XmuReleaseStippledPixmap(XtScreen(current), values.tile);
		XtReleaseGC(current, cl->xklist.graygc);
		XtReleaseGC(current, cl->xklist.revgc);
		XtReleaseGC(current, cl->xklist.normgc);
		XtReleaseGC(current, cl->xklist.whitegc);
        GetGCs(new);
        redraw = TRUE;
    }

    if ( ( cl->xklist.font != nl->xklist.font ) &&
				( cl->simple.international == False ) )
        nl->xklist.row_height = nl->xklist.font->max_bounds.ascent
	                    + nl->xklist.font->max_bounds.descent
			    + nl->xklist.row_space;

    else if ( ( cl->xklist.fontset != nl->xklist.fontset ) &&
				( cl->simple.international == True ) )
        nl->xklist.row_height = ext->max_ink_extent.height + nl->xklist.row_space;

    /* ...If the above two font(set) change checkers above both failed, check
    if row_space was altered.  If one of the above passed, row_height will
    already have been re-calculated. */

    else if ( cl->xklist.row_space != nl->xklist.row_space ) {

        if (cl->simple.international == True )
            nl->xklist.row_height = ext->max_ink_extent.height + nl->xklist.row_space;
        else
            nl->xklist.row_height = nl->xklist.font->max_bounds.ascent
	                        + nl->xklist.font->max_bounds.descent
			        + nl->xklist.row_space;
    }

    if ((cl->core.width           != nl->core.width)           ||
	(cl->core.height          != nl->core.height)          ||
	(cl->xklist.internal_width  != nl->xklist.internal_width)  ||
	(cl->xklist.internal_height != nl->xklist.internal_height) ||
	(cl->xklist.column_space    != nl->xklist.column_space)    ||
	(cl->xklist.row_space       != nl->xklist.row_space)       ||
	(cl->xklist.default_cols    != nl->xklist.default_cols)    ||
	(  (cl->xklist.force_cols   != nl->xklist.force_cols) &&
	   (rl->xklist.force_cols   != nl->xklist.ncols) )         ||
	(cl->xklist.vertical_cols   != nl->xklist.vertical_cols)   ||
	(cl->xklist.longest         != nl->xklist.longest)         ||
	(cl->xklist.nitems          != nl->xklist.nitems)          ||
	(cl->xklist.font            != nl->xklist.font)            ||
   /* Equiv. fontsets might have different values, but the same fonts, so the
   next comparison is sloppy but not dangerous.  */
	(cl->xklist.fontset         != nl->xklist.fontset)         ||
	(cl->xklist.list            != nl->xklist.list)          )   {

        CalculatedValues( new );
        Layout( new, WidthFree( nl ), HeightFree( nl ),
			 &nl->core.width, &nl->core.height );
        redraw = TRUE;
    }

    if (cl->xklist.list != nl->xklist.list)
	nl->xklist.is_highlighted = nl->xklist.highlight = NO_HIGHLIGHT;

    if ((cl->core.sensitive != nl->core.sensitive) ||
	(cl->core.ancestor_sensitive != nl->core.ancestor_sensitive)) {
        nl->xklist.highlight = NO_HIGHLIGHT;
	redraw = TRUE;
    }
    
    if (!XtIsRealized(current))
      return(FALSE);
      
    return(redraw);
}

static void Destroy(w)
    Widget w;
{
    XKListWidget lw = (XKListWidget) w;
    XGCValues values;
    
    XGetGCValues(XtDisplay(w), lw->xklist.graygc, GCTile, &values);
    XmuReleaseStippledPixmap(XtScreen(w), values.tile);
    XtReleaseGC(w, lw->xklist.graygc);
    XtReleaseGC(w, lw->xklist.revgc);
    XtReleaseGC(w, lw->xklist.normgc);
    XtReleaseGC(w, lw->xklist.whitegc);
}

// Exported Functions.

// Changes the list being used and shown.
// If nitems of longest are <= 0 then they will be calculated.
// If nitems is <= 0 then the list needs to be NULL terminated.

void XKListChange(Widget w, char **list, int nitems, int longest,
	  Boolean resize_it)
{
    XKListWidget	lw = (XKListWidget) w;
    Dimension		new_width = w->core.width;
    Dimension		new_height = w->core.height;

    lw->xklist.list = list;

    if ((nitems <= 0)) { nitems = 0; }
    lw->xklist.nitems = nitems;
    if ((longest <= 0)) { longest = 0; }

    // If the user passes 0 meaning "calculate it", it must be free.
    if ((longest != 0))
        { lw->xklist.freedoms |= LongestLock; }
    else // the user's word is god.
        { lw->xklist.freedoms &= ~LongestLock; }

    if ((resize_it))
        { lw->xklist.freedoms &= ~WidthLock & ~HeightLock; }
    // else - still resize if its not locked

    lw->xklist.longest = longest;

    CalculatedValues(w);

    if (Layout(w, WidthFree(w), HeightFree(w), &new_width, &new_height ))
        { ChangeSize(w, new_width, new_height); }

    lw->xklist.is_highlighted = lw->xklist.highlight = NO_HIGHLIGHT;
    if (XtIsRealized(w))
		{ Redisplay(w, (XEvent *) NULL, (Region) NULL); }
}

// NOTE: there's no checking on this.
void XKListSetArrows(Widget w, Boolean *arrows)
{ 
    XKListWidget	lw = (XKListWidget) w;

	lw->xklist.arrows = arrows;
    if (XtIsRealized(w))
		{ Redisplay(w, (XEvent *) NULL, (Region) NULL); }
}

// Unlights the current highlighted element.
void XKListUnhighlight(Widget w)
{
    XKListWidget	lw = (XKListWidget) w;

    lw->xklist.highlight = NO_HIGHLIGHT;
    if ((lw->xklist.is_highlighted != NO_HIGHLIGHT))
	{
        PaintItemName(w, lw->xklist.is_highlighted); /* unhighlight this one. */
	}
}

// Highlights the given item.
void XKListHighlight(Widget w, int item)
{
    XKListWidget	lw = (XKListWidget) w;
    
    if (XtIsSensitive(w))
	{
        lw->xklist.highlight = item;
        if ((lw->xklist.is_highlighted != NO_HIGHLIGHT))
		{
            PaintItemName(w, lw->xklist.is_highlighted);  /* Unhighlight. */
		}
		PaintItemName(w, item); /* HIGHLIGHT this one. */ 
    }
}

// Returns the currently highlighted object.
XKListReturnStruct *XKListShowCurrent(Widget w)
{
    XKListWidget		lw = (XKListWidget) w;
    XKListReturnStruct	*ret_val;

    ret_val = (XKListReturnStruct *) XtMalloc (sizeof (XKListReturnStruct));
    ret_val->list_index = lw->xklist.highlight;
    if ((ret_val->list_index == XK_LIST_NONE))
		{ ret_val->string = ""; }
    else
		{ ret_val->string = lw->xklist.list[ret_val->list_index]; }

    return ret_val;
}

