// XKList.h

#ifndef _XKList_h_
#define _XKList_h_

#include <X11/neXtaw/Simple.h>

/* Resources:

 Name		     	Class			RepType			Default Value
 ----		     	-----			-------			-------------
 arrows				List			Boolean*		NULL
 background	     	Background		Pixel			XtDefaultBackground
 border		     	BorderColor		Pixel			XtDefaultForeground
 borderWidth	    BorderWidth		Dimension		1
 callback           Callback        XtCallbackList  NULL       **6
 columnSpacing      Spacing         Dimension       6
 cursor		     	Cursor			Cursor			left_ptr
 cursorName	     	Cursor			String			NULL
 defaultColumns     Columns         int             2          **5
 destroyCallback    Callback		Pointer			NULL 
 font		     	Font			XFontStruct*	XtDefaultFont
 forceColumns       Columns         Boolean         False      **5
 foreground	     	Foreground		Pixel			XtDefaultForeground
 height		     	Height			Dimension		0          **1
 insensitiveBorder  Insensitive		Pixmap			Gray
 internalHeight	    Height			Dimension		2
 internalWidth	    Width			Dimension		4
 list               List            String*        NULL       **2
 longest            Longest         int             0          **3  **4
 mappedWhenManaged  MappedWhenManaged Boolean		True
 numberStrings      NumberStrings   int             0          **4
 pasteBuffer        Boolean         Boolean         False
 pointerColor	    Foreground		Pixel			XtDefaultForeground
 pointerColorBackground Background	Pixel			XtDefaultBackground
 rowSpacing         Spacing         Dimension       4
 sensitive	     	Sensitive		Boolean			True
 verticalList       Boolean         Boolean         False
 width		     	Width			Dimension		0          **1
 x		     		Position		Position		0
 y		     		Position		Position		0

 **1 - If the Width or Height of the list widget is zero (0) then the value
       is set to the minimum size necessay to fit the entire list.

       If both Width and Height are zero then they are adjusted to fit the
       entire list that is created width the number of default columns 
       specified in the defaultColumns resource.

 **2 - This is an array of strings the specify elements of the list.
       This resource must be specified. 
       (What good is a list widget without a list??  :-)

 **3 - Longest is the length of the widest string in pixels.

 **4 - If either of these values are zero (0) then the list widget calculates
       the correct value. 

       (This allows you to make startup faster if you already have 
        this information calculated)

       NOTE: If the numberStrings value is zero the list must 
             be NULL terminated.

 **5 - By setting the List.Columns resource you can force the application to
       have a given number of columns.	     
        
 **6 - This returns the name and index of the item selected in an 
       XKListReturnStruct that is pointed to by the client_data
       in the CallbackProc.
*/

// Value returned when there are no highlighted objects. 
#define XK_LIST_NONE -1	

#ifndef XtCList
#define XtCList				"List"
#endif

#ifndef XtCSpacing
#define XtCSpacing			"Spacing"
#endif

#ifndef XtCColumns
#define XtCColumns			"Columns"
#endif

#ifndef XtCLongest
#define XtCLongest			"Longest"
#endif

#ifndef XtCNumberStrings
#define XtCNumberStrings	"NumberStrings"
#endif

#ifndef XtNcursor
#define XtNcursor			"cursor"
#endif

#ifndef XtNcolumnSpacing
#define XtNcolumnSpacing	"columnSpacing"
#endif

#ifndef XtNdefaultColumns
#define XtNdefaultColumns	"defaultColumns"
#endif

#ifndef XtNforceColumns
#define XtNforceColumns		"forceColumns"
#endif

#ifndef XtNlist
#define XtNlist				"list"
#endif

#ifndef XtNlongest
#define XtNlongest			"longest"
#endif

#ifndef XtNnumberStrings
#define XtNnumberStrings	"numberStrings"
#endif

#ifndef XtNpasteBuffer
#define XtNpasteBuffer		"pasteBuffer"
#endif

#ifndef XtNrowSpacing
#define XtNrowSpacing		"rowSpacing"
#endif

#ifndef XtNverticalList
#define XtNverticalList		"verticalList"
#endif
 
#ifndef XtNfontSet
#define XtNfontSet			"fontSet"
#endif

#ifndef XtCFontSet
#define XtCFontSet			"FontSet"
#endif

// Class record constants

extern WidgetClass xkListWidgetClass;

typedef	struct	_XKListClassRec *XKListWidgetClass;
typedef	struct	_XKListRec      *XKListWidget;

// The list return structure.
typedef	struct	_XKListReturnStruct
{
	String	string;
	int		list_index;
} XKListReturnStruct;

_XFUNCPROTOBEGIN

// Changes the list being used and shown.
// If nitems of longest are <= 0 then they will be caluculated.
// If nitems is <= 0 then the list needs to be NULL terminated.
extern	void XKListChange(Widget, String*, int, int, Boolean);

// Set the arrows array.
extern	void XKListSetArrows(Widget, Boolean*);

// Unlights the current highlighted element.
extern	void XKListUnhighlight(Widget);

// Highlights the given item.
extern	void XKListHighlight(Widget, int);

// Returns the currently highlighted object.
extern	XKListReturnStruct *XKListShowCurrent(Widget);

_XFUNCPROTOEND

#endif // _XKList_h_

