//
//  Login.app
// 
//  Copyright (c) 1997, 1998 Per Liden
// 
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include "Panel.h"

 
Panel::Panel(int argc, char** argv)
{
    int tmp;
    char* dispname = NULL;

    // Parse command line
    while ((tmp = getopt(argc, argv, "d:h?")) != EOF) {
	switch (tmp) {     	    
	case 'd':	// Display
	    dispname = optarg;      
	    break;	    
	case '?':	// Ilegal
	    cerr << endl;
	case 'h':   // Help
	    cerr << "usage:  " << APPNAME << " [option ...]" << endl
		 << "options:" << endl
		 << "        -version" << endl	   
		 << "        -display  <disp>" << endl;	   
	    exit(0);
	    break;
	}
    }
     
    // Open display
    if((Dpy = XOpenDisplay(dispname)) == NULL) 
    {
	cerr << APPNAME << ": could not open display '" << dispname << "'" << endl;
	exit(1);
    }       
 
    // Get screen and root window
    Scr = DefaultScreen(Dpy);
    Root = RootWindow(Dpy, Scr);
    
    // Init GC
    XGCValues gcv;
    unsigned long gcm;           
    gcm = GCForeground|GCBackground|GCGraphicsExposures|GCFont;
    gcv.foreground = GetColor("black");
    gcv.background = GetColor("white");
    gcv.graphics_exposures = False;
    gcv.font = XLoadFont(Dpy, INPUT_FONT);
    TextGC = XCreateGC(Dpy, Root, gcm, &gcv);

    // Load background xpm   
    PanelPixmapAttributes.valuemask = 0;
    tmp = XpmCreatePixmapFromData(Dpy, Root, PanelXpm, 
				  &PanelPixmap, NULL, &PanelPixmapAttributes);
    // Check for error
    switch(tmp)
    {
    case XpmColorError:
	cerr << APPNAME << ": xpm image loaded but did not get all colors needed" << endl;
	break;

    case XpmColorFailed:
	cerr << APPNAME << ": could not load xpm image (not enough colors available)" << endl;
	exit(0);
	break;

    case XpmNoMemory:
	cerr << APPNAME << ": could not load xpm image (not enough memory available)" << endl;
	exit(0);
	break;

    case XpmOpenFailed:
    case XpmFileInvalid:
	cerr << APPNAME << ": could not load xpm image (image broken or corrupt)" << endl;
	exit(0);
	break;

    case XpmSuccess:
    default:
	// Image loaded ok
	break;
    }

    
    ShrinkPixmap = XCreatePixmap(Dpy, Root, 
				 PanelPixmapAttributes.width, 
				 PanelPixmapAttributes.height,
				 DefaultDepth(Dpy, Scr));

    // Calculate position
    X = (XWidthOfScreen(ScreenOfDisplay(Dpy, Scr)) - PanelPixmapAttributes.width) / 2;
    Y = (XHeightOfScreen(ScreenOfDisplay(Dpy, Scr)) - PanelPixmapAttributes.height - 100) / 2;
  
    // Init In
    In = new Input;
}


Panel::~Panel()
{    
    // Never called
}


void Panel::Run()
{
    XEvent event;
    int panelclosed = 1;

    // Start looping
    while(1) 
    {	
	if(panelclosed)	
	    OpenPanel();

	Action = WAIT;
	In->Reset();

	while(Action == WAIT)
	{
	    XNextEvent(Dpy, &event);
	    EventHandler(&event);	    
	}	

	if(Action == FAIL)
	{
	    panelclosed = 0;
	    ShakePanel();	   
	}
	else
	{
	    panelclosed = 1;
	    ClosePanel();
	    
	    switch(Action)
	    {
	    case LOGIN: Login(); break;		
	    case CONSOLE: Console(); break;		
	    case REBOOT: Reboot(); break;
	    case HALT: Halt(); break;
	    case EXIT: Exit(); break;
	    }
	}
    }
}


void Panel::OpenPanel()
{
    // Init root
    system(INIT_CMD);

    // Close all clients
    KillAllClients(False);
    KillAllClients(True);

    // Create window
    Win = XCreateSimpleWindow(Dpy, Root, X, Y, 
			      PanelPixmapAttributes.width,
			      PanelPixmapAttributes.height, 
			      1, GetColor("black"), GetColor("white"));
    
    // Events
    XSelectInput(Dpy, Win, ExposureMask | KeyPressMask);

    // Set background
    XSetWindowBackgroundPixmap(Dpy, Win, PanelPixmap);

    // Show window
    XMapWindow(Dpy, Win);
 
    // Grab keyboard
    XGrabKeyboard(Dpy, Win, False, GrabModeAsync, GrabModeAsync, CurrentTime);
	
    XFlush(Dpy);
}


void Panel::ClosePanel()
{
    ShrinkPanel();
    XUngrabKeyboard(Dpy, CurrentTime);
    XUnmapWindow(Dpy, Win);
    XDestroyWindow(Dpy, Win);
    XFlush(Dpy);
}


void Panel::ShakePanel()
{
    // Erase name and passwd  
    char* name = In->GetName();
    char* passwd = In->GetHiddenPasswd();
    Cursor(HIDE);
    XSetForeground(Dpy, TextGC, GetColor(INPUT_BGCOLOR));
    XDrawString(Dpy, Win, TextGC, INPUT_NAME_X, INPUT_NAME_Y, name, strlen(name));
    XDrawString(Dpy, Win, TextGC, INPUT_PASSWD_X, INPUT_PASSWD_Y, passwd, strlen(passwd));
    In->Reset();
    Cursor(SHOW);

    // Shake panel
    for(int i=0;i<3;i++)
    {
        XMoveWindow(Dpy, Win, X+2, Y);       
        XMoveWindow(Dpy, Win, X+4, Y);       
        XMoveWindow(Dpy, Win, X+6, Y);       
        XMoveWindow(Dpy, Win, X+8, Y);       
        XMoveWindow(Dpy, Win, X+10, Y);
        XMoveWindow(Dpy, Win, X+12, Y);       
        XMoveWindow(Dpy, Win, X+14, Y);       
        XMoveWindow(Dpy, Win, X+16, Y);       
        XMoveWindow(Dpy, Win, X+18, Y);       
        XMoveWindow(Dpy, Win, X+20, Y);
        XMoveWindow(Dpy, Win, X+18, Y);       
        XMoveWindow(Dpy, Win, X+16, Y);       
        XMoveWindow(Dpy, Win, X+14, Y);       
        XMoveWindow(Dpy, Win, X+12, Y);       
        XMoveWindow(Dpy, Win, X+10, Y);              
        XMoveWindow(Dpy, Win, X+8, Y);
        XMoveWindow(Dpy, Win, X+6, Y);       
        XMoveWindow(Dpy, Win, X+4, Y);
        XMoveWindow(Dpy, Win, X+2, Y);       
        XMoveWindow(Dpy, Win, X, Y);	
        XMoveWindow(Dpy, Win, X-2, Y);       
        XMoveWindow(Dpy, Win, X-4, Y);       
        XMoveWindow(Dpy, Win, X-6, Y);       
        XMoveWindow(Dpy, Win, X-8, Y);       
        XMoveWindow(Dpy, Win, X-10, Y);
        XMoveWindow(Dpy, Win, X-12, Y);       
        XMoveWindow(Dpy, Win, X-14, Y);       
        XMoveWindow(Dpy, Win, X-16, Y);       
        XMoveWindow(Dpy, Win, X-18, Y);       
        XMoveWindow(Dpy, Win, X-20, Y);
        XMoveWindow(Dpy, Win, X-18, Y);       
        XMoveWindow(Dpy, Win, X-16, Y);       
        XMoveWindow(Dpy, Win, X-14, Y);       
        XMoveWindow(Dpy, Win, X-12, Y);       
        XMoveWindow(Dpy, Win, X-10, Y);              
        XMoveWindow(Dpy, Win, X-8, Y);
        XMoveWindow(Dpy, Win, X-6, Y);       
        XMoveWindow(Dpy, Win, X-4, Y);
        XMoveWindow(Dpy, Win, X-2, Y);       
        XMoveWindow(Dpy, Win, X, Y);	
	XFlush(Dpy);
    }
}


void Panel::ShrinkPanel()
{
    int x = X;
    int y = Y;
    int width = PanelPixmapAttributes.width;
    int height = PanelPixmapAttributes.height; 

    XCopyArea(Dpy, PanelPixmap, ShrinkPixmap, TextGC,
 	      0, 0, width, height, 0, 0);

    for(;width > 0; width-=10, x+=5)
    {
	XMoveResizeWindow(Dpy, Win, x, y, width, height);
	XCopyArea(Dpy, ShrinkPixmap, ShrinkPixmap, TextGC, 
		  5, 0, width, height, 0, 0);
	XSetWindowBackgroundPixmap(Dpy, Win, ShrinkPixmap);
	XFlush(Dpy);
    }
    
    XSetWindowBackgroundPixmap(Dpy, Win, PanelPixmap);
}


void Panel::Login()
{
    struct passwd *pw;
    pid_t pid;

    pw = In->GetPasswdStruct();
    if(pw == 0)
	return;
	
    // Create new process
    pid = fork();
    if(pid == 0)
    {
	// Login process starts here
	SwitchUser Su(APPNAME, pw);
        Su.Login(LOGIN_CMD);
	exit(0);
    }

    // Wait until user is logging out (login process terminates)
    waitpid(pid, 0, 0);

    // Close all clients
    KillAllClients(False);
    KillAllClients(True);

    // Send HUP signal to clientgroup
    killpg(pid, SIGHUP);

    // Send TERM signal to clientgroup, if error send KILL
    if(killpg(pid, SIGTERM))
	killpg(pid, SIGKILL);
}


void Panel::Reboot()
{
    Message("System rebooting...");
    system(REBOOT_CMD);
    sleep(60);
}


void Panel::Halt()
{
    Message("System halting...");
    system(HALT_CMD);
    sleep(60);
}


void Panel::Console()
{
    int posx = 40;
    int posy = 40;
    int fontx = 9;
    int fonty = 15;
    int width = (XWidthOfScreen(ScreenOfDisplay(Dpy, Scr)) - (posx * 2)) / fontx;
    int height = (XHeightOfScreen(ScreenOfDisplay(Dpy, Scr)) - (posy * 2)) / fonty;

    char *tmp = new char[strlen(CONSOLE_CMD) + 60];
    sprintf(tmp, CONSOLE_CMD, width, height, posx, posy, fontx, fonty);
    system(tmp);
    delete [] tmp;
}


void Panel::Exit()
{
    XFreePixmap(Dpy, ShrinkPixmap);
    XFreeGC(Dpy, TextGC);
    XCloseDisplay(Dpy);
    delete In;
    exit(0);
}


void Panel::Message(char* text)
{
    XSetForeground(Dpy, TextGC, GetColor(MSG_COLOR));
    XDrawString(Dpy, Root, TextGC, 10, 25, text, strlen(text));
    XFlush(Dpy);
}


unsigned long Panel::GetColor(char* colorname)
{
    XColor color;
    XWindowAttributes attributes;
    
    XGetWindowAttributes(Dpy, Root, &attributes);
    color.pixel = 0;
    
    if(!XParseColor(Dpy, attributes.colormap, colorname, &color)) 
	cerr << APPNAME << ": can't parse color " << colorname << endl;
    else if(!XAllocColor(Dpy, attributes.colormap, &color)) 
	cerr << APPNAME << ": can't allocate color " << colorname << endl;   
    
    return color.pixel;
}


int CatchErrors(Display *dpy, XErrorEvent *ev)
{
    return 0;
}


void Panel::KillAllClients(Bool top)
{
    Window dummywindow;
    Window *children;
    unsigned int nchildren;
    unsigned int i;
    XWindowAttributes attr;

    XSync(Dpy, 0);
    XSetErrorHandler(CatchErrors);

    nchildren = 0;
    XQueryTree(Dpy, Root, &dummywindow, &dummywindow, &children, &nchildren);
    if(!top) 
    {
	for(i=0; i<nchildren; i++) 
	{
	    if(XGetWindowAttributes(Dpy, children[i], &attr) && (attr.map_state == IsViewable))
		children[i] = XmuClientWindow(Dpy, children[i]);
	    else
		children[i] = 0;
	}
    }

    for(i=0; i<nchildren; i++) 
    {
	if(children[i])
	    XKillClient(Dpy, children[i]);
    }
    XFree((char *)children);
    
    XSync(Dpy, 0);
    XSetErrorHandler(NULL);    
}


void Panel::Cursor(int visible)
{       
    char* text;
    int xx, yy;

    switch(In->GetField())
    {
    case GET_PASSWD:
	text = In->GetHiddenPasswd();
	xx = INPUT_PASSWD_X;
	yy = INPUT_PASSWD_Y;
	break;

    case GET_NAME:
	text = In->GetName();
	xx = INPUT_NAME_X;
	yy = INPUT_NAME_Y;
	break;
    }

    if(visible == SHOW)
	XSetForeground(Dpy, TextGC, GetColor(INPUT_FGCOLOR));
    else
	XSetForeground(Dpy, TextGC, GetColor(INPUT_BGCOLOR));
    
    xx += XTextWidth(XLoadQueryFont(Dpy, INPUT_FONT), text, strlen(text));    
    XDrawLine(Dpy, Win, TextGC, 
	      xx, yy-(INPUT_CURSOR_HEIGHT-3), 
	      xx, yy+3);
    XDrawLine(Dpy, Win, TextGC, 
	      xx+1, yy-(INPUT_CURSOR_HEIGHT-3), 
	      xx+1, yy+3); 
}


void Panel::EventHandler(XEvent* event)
{
    switch(event->type) 
    {	
    case Expose:
	OnExpose(event);
	break;
	
    case KeyPress:
	OnKeyPress(event);
	break;
    }
}


void Panel::OnExpose(XEvent* event)
{
    char* name = In->GetName();
    char* passwd = In->GetHiddenPasswd();
    XSetForeground(Dpy, TextGC, GetColor(INPUT_FGCOLOR));
    XDrawString(Dpy, Win, TextGC, INPUT_NAME_X, INPUT_NAME_Y, name, strlen(name));
    XDrawString(Dpy, Win, TextGC, INPUT_PASSWD_X, INPUT_PASSWD_Y, passwd, strlen(passwd));
    Cursor(SHOW);
}
 

void Panel::OnKeyPress(XEvent* event)
{

    char del;
    char buffer;    
    KeySym keysym;
    XComposeStatus compstatus;
    int xx;
    int yy;
    char* text;

    Cursor(HIDE);
    XLookupString(&event->xkey, &buffer, 1, &keysym, &compstatus);
    del = In->Key(buffer, keysym);
    Action = In->GetAction();

    switch(In->GetField())
    {
    case GET_PASSWD:
	text = In->GetHiddenPasswd();
	xx = INPUT_PASSWD_X;
	yy = INPUT_PASSWD_Y;
	break;

    case GET_NAME:
	text = In->GetName();
	xx = INPUT_NAME_X;
	yy = INPUT_NAME_Y;
	break;	
    }

    
    if(del == 0)
    {
	// No character deleted
	XSetForeground(Dpy, TextGC, GetColor(INPUT_FGCOLOR));
	XDrawString(Dpy, Win, TextGC, xx, yy, text, strlen(text));
    }
    else
    {
	// Character deleted
	int i = 0;
	char tmp[2];
	tmp[0] = del;
	tmp[1] = '\0';

	XSetForeground(Dpy, TextGC, GetColor(INPUT_BGCOLOR));
	i = XTextWidth(XLoadQueryFont(Dpy, INPUT_FONT), text, strlen(text));
	XDrawString(Dpy, Win, TextGC, xx + i, yy, tmp, 1);
    }

    Cursor(SHOW);
}

