/* $XConsortium: Login.c,v 1.42 94/04/17 20:03:53 gildea Exp $ */
/*

Copyright (c) 1988  X Consortium

Permission is hereby granted, free of charge, to any person obtaining
a copy of this software and associated documentation files (the
"Software"), to deal in the Software without restriction, including
without limitation the rights to use, copy, modify, merge, publish,
distribute, sublicense, and/or sell copies of the Software, and to
permit persons to whom the Software is furnished to do so, subject to
the following conditions:

The above copyright notice and this permission notice shall be included
in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
IN NO EVENT SHALL THE X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR
OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall
not be used in advertising or otherwise to promote the sale, use or
other dealings in this Software without prior written authorization
from the X Consortium.

*/

/*
	xdm		-	display manager daemon
	Author:		Keith Packard, MIT X Consortium

	Login.c

	xdm		-	AfterStep/OpenStep login widget
	Author:		mitch@redback.com.au
*/

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/keysym.h>
#include <X11/Xfuncs.h>

#include <stdio.h>

#include <dm.h>
#include <greet.h>
#include <LoginP.h>

#ifdef GNUSTEP
#include <GNUSTEP.xpm>
#else
#include <GNULinux.xpm>
#endif

#define offset(field) XtOffsetOf(LoginRec, login.field)
#define goffset(field) XtOffsetOf(WidgetRec, core.field)

extern Debug(), LogOutOfMem();
int	ResetLogin(LoginWidget);

static XtResource resources[] =
{
    { XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
		goffset(width), XtRImmediate, (XtPointer) 0 },
    { XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
		goffset(height), XtRImmediate, (XtPointer) 0 },
    { XtNx, XtCX, XtRPosition, sizeof(Position),
		goffset(x), XtRImmediate, (XtPointer) -1 },
    { XtNy, XtCY, XtRPosition, sizeof(Position),
		goffset(y), XtRImmediate, (XtPointer) -1 },
    { XtNborderWidth, XtCBorderWidth, XtRDimension, sizeof(Position),
		goffset(border_width), XtRImmediate, (XtPointer) 0 },

    { XtNforeground, XtCForeground, XtRPixel, sizeof(Pixel),
        offset(textpixel), XtRString, XtDefaultForeground},
    { XtNfont, XtCFont, XtRFontStruct, sizeof(XFontStruct *),
    	offset(font), XtRString, "*-helvetica-medium-r-normal-*-180-*" },
    { XtNfailTimeout, XtCFailTimeout, XtRInt, sizeof(int),
		offset(failTimeout), XtRImmediate, (XtPointer) 10 },
    { XtNnotifyDone, XtCCallback, XtRFunction, sizeof(XtPointer),
		offset(notify_done), XtRFunction, (XtPointer) 0 },
    { XtNsessionArgument, XtCSessionArgument, XtRString, sizeof(char *),
		offset(sessionArg), XtRString, (XtPointer) 0 },
    { XtNsecureSession, XtCSecureSession, XtRBoolean, sizeof(Boolean),
		offset(secure_session), XtRImmediate, False },
    { XtNallowAccess, XtCAllowAccess, XtRBoolean, sizeof(Boolean),
		offset(allow_access), XtRImmediate, False }
};

#undef offset
#undef goffset

// Input coordinates.
#define LOGIN_X		200
#define LOGIN_Y		154
#define PASSWD_X	200
#define PASSWD_Y	200

static	void	Initialize();
static	void	Realize();
static	void	Destroy();
static	void	Redisplay();
static	Boolean	SetValues();
static	void	DrawIt();

// Convenience.
GC GetGC(Widget w)
{
	return ((LoginWidget) w)->login.textGC;
}

// Adding textGbGC for white background. w->login.textBgGC for erase,
// w->login.textGC for drawing.
static void RenderField(LoginWidget w, GC currGC, int cursor, char *str,
	int yoffset)
{
    int	x;

    x = LOGIN_X;
    if (cursor > 0) { x += XTextWidth(w->login.font, str, cursor); }

    XDrawString(XtDisplay(w),
		XtWindow(w),
		currGC,
		x,
		yoffset,
		str + cursor,
		strlen(str + cursor));
}

static void RealizeCursor(LoginWidget w, GC gc)
{
    int		x;
	int		y;
    int		height;
	int		width;
	char	*str;

    switch (w->login.state)
	{
		case GET_NAME:
			x = LOGIN_X;
			y = LOGIN_Y;
			str = w->login.data.name;
			break;
		case GET_PASSWD:
			x = PASSWD_X;
			y = PASSWD_Y;
			// Adding for cursor movement in password field.
			str = w->login.data.bogus;
			break;
		default:
			return;
	}

	if (w->login.cursor > 0)
		{ x += XTextWidth(w->login.font, str, w->login.cursor); }

	// Extracted these from above. Merely repetitious.
	height = w->login.font->max_bounds.ascent +
		w->login.font->max_bounds.descent;
	width = 1;

	// Draw cursor?
    XFillRectangle(XtDisplay(w),
		XtWindow(w),
		gc,
		x,
		y + 1 - w->login.font->max_bounds.ascent,
		width,
		height - 1);
    XDrawPoint(XtDisplay(w),
		XtWindow(w),
		gc,
		x - 1,
		y - w->login.font->max_bounds.ascent);
    XDrawPoint(XtDisplay(w),
		XtWindow(w),
		gc,
		x + 1,
		y - w->login.font->max_bounds.ascent);
    XDrawPoint(XtDisplay(w),
		XtWindow(w),
		gc,
		x - 1,
		y - w->login.font->max_bounds.ascent + height);
    XDrawPoint(XtDisplay(w),
		XtWindow(w),
		gc,
		x + 1,
		y - w->login.font->max_bounds.ascent + height);
    XDrawPoint(XtDisplay(w),
		XtWindow(w),
		gc,
		x - 2,
		y - w->login.font->max_bounds.ascent);
    XDrawPoint(XtDisplay(w),
		XtWindow(w),
		gc,
		x + 2,
		y - w->login.font->max_bounds.ascent);
    XDrawPoint(XtDisplay(w),
		XtWindow(w),
		gc,
		x - 2,
		y - w->login.font->max_bounds.ascent + height);
    XDrawPoint(XtDisplay(w),
		XtWindow(w),
		gc,
		x + 2,
		y - w->login.font->max_bounds.ascent + height);
}

static void EraseFail(LoginWidget w) { }

static void XorCursor(LoginWidget w) { RealizeCursor(w, w->login.xorGC); }

static void RemoveFail(LoginWidget w) { if (w->login.failUp) { EraseFail(w); } }

static void EraseCursor(LoginWidget w) { RealizeCursor(w, w->login.textBgGC); }

void failTimeout(XtPointer client_data, XtIntervalId *id)
{
    LoginWidget	w = (LoginWidget) client_data;

    Debug("failTimeout\n");
    EraseFail(w);
}

int DrawFail(Widget ctx)
{
    LoginWidget	w = (LoginWidget) ctx;
	Display		*dpy = XtDisplay(ctx);
	int			i;
	Position	X = w->core.x;
	Position	Y = w->core.y;

    // Shake panel
    for (i = 0; i < 3; i++)
    {
        XtMoveWidget(ctx, X + 2, Y);       
        XtMoveWidget(ctx, X + 4, Y);       
        XtMoveWidget(ctx, X + 6, Y);       
        XtMoveWidget(ctx, X + 8, Y);       
        XtMoveWidget(ctx, X + 10, Y);
        XtMoveWidget(ctx, X + 12, Y);       
        XtMoveWidget(ctx, X + 14, Y);       
        XtMoveWidget(ctx, X + 16, Y);       
        XtMoveWidget(ctx, X + 18, Y);       
        XtMoveWidget(ctx, X + 20, Y);
        XtMoveWidget(ctx, X + 18, Y);       
        XtMoveWidget(ctx, X + 16, Y);       
        XtMoveWidget(ctx, X + 14, Y);       
        XtMoveWidget(ctx, X + 12, Y);       
        XtMoveWidget(ctx, X + 10, Y);              
        XtMoveWidget(ctx, X + 8, Y);
        XtMoveWidget(ctx, X + 6, Y);       
        XtMoveWidget(ctx, X + 4, Y);
        XtMoveWidget(ctx, X + 2, Y);       
        XtMoveWidget(ctx, X, Y);	
        XtMoveWidget(ctx, X - 2, Y);       
        XtMoveWidget(ctx, X - 4, Y);       
        XtMoveWidget(ctx, X - 6, Y);       
        XtMoveWidget(ctx, X - 8, Y);       
        XtMoveWidget(ctx, X - 10, Y);
        XtMoveWidget(ctx, X - 12, Y);       
        XtMoveWidget(ctx, X - 14, Y);       
        XtMoveWidget(ctx, X - 16, Y);       
        XtMoveWidget(ctx, X - 18, Y);       
        XtMoveWidget(ctx, X - 20, Y);
        XtMoveWidget(ctx, X - 18, Y);       
        XtMoveWidget(ctx, X - 16, Y);       
        XtMoveWidget(ctx, X - 14, Y);       
        XtMoveWidget(ctx, X - 12, Y);       
        XtMoveWidget(ctx, X - 10, Y);              
        XtMoveWidget(ctx, X - 8, Y);
        XtMoveWidget(ctx, X - 6, Y);       
        XtMoveWidget(ctx, X - 4, Y);
        XtMoveWidget(ctx, X - 2, Y);       
        XtMoveWidget(ctx, X, Y);	
		XFlush(dpy);
    }

    XorCursor(w);
    ResetLogin(w);
    XorCursor(w);
    w->login.failUp = 1;

    if (w->login.failTimeout > 0)
	{
		Debug("failTimeout: %d\n", w->login.failTimeout);
		XtAppAddTimeOut(XtWidgetToApplicationContext((Widget)w),
			w->login.failTimeout * 1000,
			failTimeout,
			(XtPointer) w);
    }
	return 0;
}

static void DrawIt(LoginWidget w)
{
    EraseCursor(w);

    RenderField(w, w->login.textGC, 0, w->login.data.name, LOGIN_Y);
    XorCursor(w);

	// The GrabKeyboard here is needed only because of a bug in the R3 
	// server -- the keyboard is grabbed on the root window, and the 
	// server won't dispatch events to the focus window unless the focus 
	// window is a ancestor of the grab window. Bug in server already 
	// found and fixed, compatibility until at least R4.

    if (XGrabKeyboard (XtDisplay(w), XtWindow(w), False, GrabModeAsync,
		       GrabModeAsync, CurrentTime) != GrabSuccess)
    {
		XSetInputFocus(XtDisplay(w),
			XtWindow(w),
			RevertToPointerRoot,
			CurrentTime);
    }
}

static void DeleteBackwardChar(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;
	char		*str;
	int			yoffset;

    XorCursor(ctx);
    RemoveFail(ctx);

    if (ctx->login.cursor > 0)
	{
		ctx->login.cursor--;
		switch (ctx->login.state)
		{
			case GET_NAME:
				str = ctx->login.data.name;
				yoffset = LOGIN_Y;
				break;
			case GET_PASSWD:
				strcpy(ctx->login.data.passwd + ctx->login.cursor,
					ctx->login.data.passwd + ctx->login.cursor + 1);
				str = ctx->login.data.bogus;
				yoffset = PASSWD_Y;
				break;
		}

		// Assuming login.state can only have these 2 values. Adding for '*'.
		RenderField(ctx,
			ctx->login.textBgGC,
			ctx->login.cursor,
			str,
			yoffset);
		strcpy(str + ctx->login.cursor, str + ctx->login.cursor + 1);
		RenderField(ctx,
			ctx->login.textGC,
			ctx->login.cursor,
			str,
			yoffset);
    }
    XorCursor(ctx);	
}

static void DeleteForwardChar(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;
	char		*str;
	int			yoffset;

    XorCursor(ctx);
    RemoveFail(ctx);

    switch (ctx->login.state)
	{
		case GET_NAME:
			if (ctx->login.cursor < (int) strlen(ctx->login.data.name))
			{
				str = ctx->login.data.name;
				yoffset = PASSWD_Y;
			}
			break;
		case GET_PASSWD:
			if (ctx->login.cursor < (int) strlen(ctx->login.data.passwd))
			{
				strcpy(ctx->login.data.passwd + ctx->login.cursor,
					ctx->login.data.passwd + ctx->login.cursor + 1);
				// Adding for '*'.
				str = ctx->login.data.bogus;
				yoffset = PASSWD_Y;
			}
			break;
    }

	if (ctx->login.cursor < (int) strlen (str))
	{
		RenderField(ctx,
			ctx->login.textBgGC,
			ctx->login.cursor,
			str,
			yoffset);
		strcpy(str + ctx->login.cursor, str + ctx->login.cursor + 1);
		RenderField(ctx,
			ctx->login.textGC,
			ctx->login.cursor,
			str,
			yoffset);
	}
    XorCursor(ctx);	
}

static void MoveBackwardChar(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget	ctx = (LoginWidget) ctxw;

    XorCursor(ctx);
    RemoveFail(ctx);
    if (ctx->login.cursor > 0)
		{ ctx->login.cursor--; }
    XorCursor(ctx);
}

static void MoveForwardChar(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;

    XorCursor(ctx);
    RemoveFail(ctx);

    switch (ctx->login.state)
	{
		case GET_NAME:
			if (ctx->login.cursor < (int) strlen(ctx->login.data.name))
				{ ++ctx->login.cursor; }
			break;
		case GET_PASSWD:
			if (ctx->login.cursor < (int) strlen(ctx->login.data.passwd))
				{ ++ctx->login.cursor; }
		break;
    }
    XorCursor(ctx);
}

static void MoveToBegining(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;

    XorCursor(ctx);
    RemoveFail(ctx);
    ctx->login.cursor = 0;
    XorCursor(ctx);
}

static void MoveToEnd(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;

    XorCursor(ctx);
    RemoveFail(ctx);

    switch (ctx->login.state)
	{
		case GET_NAME:
			ctx->login.cursor = strlen(ctx->login.data.name);
			break;
		case GET_PASSWD:
			ctx->login.cursor = strlen(ctx->login.data.passwd);
			break;
    }
    XorCursor(ctx);
}

static void EraseToEndOfLine(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;

    XorCursor(ctx);
    RemoveFail(ctx);

    switch (ctx->login.state)
	{
		case GET_NAME:
			RenderField(ctx, ctx->login.textBgGC, ctx->login.cursor,
				ctx->login.data.name, LOGIN_Y);
			ctx->login.data.name[ctx->login.cursor] = '\0';
			break;
		case GET_PASSWD:
			ctx->login.data.passwd[ctx->login.cursor] = '\0';

			// Adding for '*'.
			RenderField(ctx, ctx->login.textBgGC, ctx->login.cursor,
				ctx->login.data.bogus, PASSWD_Y);
			ctx->login.data.bogus[ctx->login.cursor] = '\0';
			break;
    }
    XorCursor(ctx);
}

static void EraseLine(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    MoveToBegining(ctxw, event, params, num_params);
    EraseToEndOfLine(ctxw, event, params, num_params);
}

static void FinishField(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;

    XorCursor(ctx);
    RemoveFail(ctx);
    switch (ctx->login.state)
	{
		case GET_NAME:
			ctx->login.state = GET_PASSWD;
			ctx->login.cursor = 0;
			break;
		case GET_PASSWD:
			ctx->login.state = DONE;
			ctx->login.cursor = 0;
			(*ctx->login.notify_done)(ctx, &ctx->login.data, NOTIFY_OK);
			break;
    }
    XorCursor(ctx);
}

static void AllowAccess(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget	ctx = (LoginWidget) ctxw;
    Arg			arglist[1];
    Boolean		allow;

    RemoveFail(ctx);
    XtSetArg(arglist[0], XtNallowAccess, (char *) &allow);
    XtGetValues((Widget) ctx, arglist, 1);
    XtSetArg(arglist[0], XtNallowAccess, !allow);
    XtSetValues((Widget) ctx, arglist, 1);
}

static void SetSessionArgument(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;

    RemoveFail(ctx);
    if (ctx->login.sessionArg)
		{ XtFree(ctx->login.sessionArg); }
    ctx->login.sessionArg = 0;

    if (*num_params > 0)
	{
		ctx->login.sessionArg = XtMalloc(strlen (params[0]) + 1);
		if (ctx->login.sessionArg)
	    	{ strcpy(ctx->login.sessionArg, params[0]); }
		else
			{ LogOutOfMem("set session argument"); }
    }
}

static void RestartSession(Widget ctxw, XEvent *event, String *params,
	Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;

    XorCursor(ctx);
    RemoveFail(ctx);
    ctx->login.state = DONE;
    ctx->login.cursor = 0;
    (*ctx->login.notify_done)(ctx, &ctx->login.data, NOTIFY_RESTART);
    XorCursor(ctx);
}

static void AbortSession(Widget ctxw, XEvent *event, String *params,
    Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;

    XorCursor(ctx);
    RemoveFail(ctx);
    ctx->login.state = DONE;
    ctx->login.cursor = 0;
    (*ctx->login.notify_done)(ctx, &ctx->login.data, NOTIFY_ABORT);
    XorCursor(ctx);
}

static void AbortDisplay(Widget ctxw, XEvent *event, String *params,
    Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;

    XorCursor(ctx);
    RemoveFail(ctx);
    ctx->login.state = DONE;
    ctx->login.cursor = 0;
    (*ctx->login.notify_done)(ctx, &ctx->login.data, NOTIFY_ABORT_DISPLAY);
    XorCursor(ctx);
}

int ResetLogin(LoginWidget w)
{
	RenderField(w, w->login.textBgGC, 0, w->login.data.name, LOGIN_Y);
    w->login.cursor = 0;
    w->login.data.name[0] = '\0';
    w->login.data.passwd[0] = '\0';

	// Adding for '*'.
	RenderField(w, w->login.textBgGC, 0, w->login.data.bogus, PASSWD_Y);
    w->login.data.bogus[0] = '\0';
    w->login.state = GET_NAME;
    return 0;
}

static void InsertChar(Widget ctxw, XEvent *event, String *params,
    Cardinal *num_params)
{
    LoginWidget ctx = (LoginWidget) ctxw;
    char		strbuf[128];
    int 		len;
	int			pixels;

    len = XLookupString(&event->xkey, strbuf, sizeof (strbuf), 0, 0);
    strbuf[len] = '\0';

    pixels = 3 + ctx->login.font->max_bounds.width * len +
		 XTextWidth(ctx->login.font,
			ctx->login.data.name,
			strlen(ctx->login.data.name));

	// Pixels to be added.
    switch (ctx->login.state)
	{
		case GET_NAME:
			if ((len + (int) strlen(ctx->login.data.name) >= NAME_LEN - 1))
				{ len = NAME_LEN - strlen(ctx->login.data.name) - 2; }
		case GET_PASSWD:
			if (len + (int) strlen(ctx->login.data.passwd) >= NAME_LEN - 1)
				{ len = NAME_LEN - strlen(ctx->login.data.passwd) - 2; }
    }

    // if (len == 0 || pixels >= LOGIN_W(ctx) - PROMPT_W(ctx)) { return; }

    XorCursor(ctx);
    RemoveFail(ctx);

    switch (ctx->login.state)
	{
		case GET_NAME:
			RenderField(ctx, ctx->login.textBgGC, ctx->login.cursor,
				ctx->login.data.name, LOGIN_Y);
			ctx->login.data.name[ctx->login.cursor] = '\0';
			memmove(ctx->login.data.name + ctx->login.cursor + len,
			   ctx->login.data.name + ctx->login.cursor,
			   strlen(ctx->login.data.name + ctx->login.cursor) + 1);
			memmove(ctx->login.data.name + ctx->login.cursor, strbuf, len);
			RenderField(ctx, ctx->login.textGC, ctx->login.cursor,
				ctx->login.data.name, LOGIN_Y);
			ctx->login.cursor += len;
			break;
		case GET_PASSWD:
			memmove(ctx->login.data.passwd + ctx->login.cursor + len,
			   ctx->login.data.passwd + ctx->login.cursor,
			   strlen(ctx->login.data.passwd + ctx->login.cursor) + 1);
			memmove(ctx->login.data.passwd + ctx->login.cursor, strbuf, len);

			// Adding for '*'.
			RenderField(ctx, ctx->login.textBgGC, ctx->login.cursor,
				ctx->login.data.bogus, PASSWD_Y);
			memmove(ctx->login.data.bogus + ctx->login.cursor + len,
			   ctx->login.data.bogus + ctx->login.cursor,
			   strlen(ctx->login.data.bogus + ctx->login.cursor) + 1);
			memset(strbuf, '*', len);
			strbuf[len] = '\0';
			memmove(ctx->login.data.bogus + ctx->login.cursor, strbuf, len);
			RenderField(ctx, ctx->login.textGC, ctx->login.cursor,
				ctx->login.data.bogus, PASSWD_Y);
			ctx->login.cursor += len;
			break;
    }
    XorCursor(ctx);
}

static void Initialize(Widget greq, Widget gnew, ArgList args,
    Cardinal *num_args)
{
    LoginWidget w = (LoginWidget) gnew;
	Display		*dpy = XtDisplay(w);
	int			screen = DefaultScreen(dpy);
    XtGCMask	valuemask, xvaluemask;
    XGCValues	myXGCV;
    Arg			wargs[10];
    Position	x;
	Position	y;

    myXGCV.foreground = BlackPixel(dpy, screen);
    myXGCV.background = WhitePixel(dpy, screen);
    myXGCV.graphics_exposures = False;
    valuemask = GCForeground | GCBackground | GCGraphicsExposures | GCFont;
	myXGCV.font = w->login.font->fid;
    w->login.textGC = XtGetGC(gnew, valuemask, &myXGCV);

	// Adding for text background.
	myXGCV.foreground = WhitePixel(dpy, screen);
	w->login.textBgGC = XtGetGC(gnew, valuemask, &myXGCV);

    myXGCV.foreground = w->login.textpixel ^ w->core.background_pixel;
    myXGCV.function = GXxor;
    xvaluemask = valuemask | GCFunction;
    w->login.xorGC = XtGetGC(gnew, xvaluemask, &myXGCV);

	// Load pixmap.
	w->login.xpma.valuemask = 0;
	switch (XpmCreatePixmapFromData(XtDisplay(w),
			DefaultRootWindow(XtDisplay(w)),
			PanelXpm,
			&w->login.pixmap,
			NULL,
			&w->login.xpma))
    {
		case XpmColorError:
			fprintf(stderr,
				": xpm image loaded but did not get all colors needed\n");
			break;
		case XpmColorFailed:
			fprintf(stderr,
				": could not load xpm image (not enough colors available)\n");
			exit(0);
			break;
		case XpmNoMemory:
			fprintf(stderr,
				": could not load xpm image (insufficient memory)\n");
			exit(0);
			break;
		case XpmOpenFailed:
		case XpmFileInvalid:
			fprintf(stderr,
				": could not load xpm image (image broken or corrupt)\n");
			exit(0);
			break;
		case XpmSuccess:
		default:
			// Image loaded ok
			break;
    }

	// Note that the second argument is a GCid -- QueryFont accepts a 
	// GCid and returns the currently contained font.
    if (!(w->login.font))
	{
		w->login.font = XQueryFont(XtDisplay(w),
			XGContextFromGC(XDefaultGCOfScreen(XtScreen(w))));
	}

	// Set sensible defaults.
    w->login.data.name[0] = '\0';
    w->login.data.passwd[0] = '\0';
	w->login.data.bogus[0] = '\0';
    w->login.state = GET_NAME;
    w->login.cursor = 0;
    w->login.failUp = 0;

	// Set geometry.
    w->core.width = w->login.xpma.width;
    w->core.height = w->login.xpma.height;
	x = (int) (XWidthOfScreen(XtScreen(w)) - w->core.width) / 2;
	y = (int) (XHeightOfScreen(XtScreen(w)) - w->core.height - 100) / 2;
    XtSetArg(wargs[0], XtNx, (XtArgVal) x);
    XtSetArg(wargs[1], XtNy, (XtArgVal) y);
	XtSetArg(wargs[2], XtNwidth, (XtArgVal) w->core.width);
	XtSetArg(wargs[3], XtNheight, (XtArgVal) w->core.height);
    XtSetValues(XtParent(w), wargs, (Cardinal) 4);
	XtVaSetValues((Widget) w,
		XtNbackgroundPixmap, (XtArgVal) w->login.pixmap,
		NULL);
}

static void Realize(Widget gw, XtValueMask *valueMask,
	XSetWindowAttributes *attrs)
{
    XtCreateWindow(gw,
		(unsigned) InputOutput,
		(Visual *) CopyFromParent,
		*valueMask,
		attrs);
}

static void Destroy(Widget gw)
{
    LoginWidget	w = (LoginWidget) gw;
	Display		*dpy = XtDisplay(w);
    Position	x = w->core.x;
    Position	y = w->core.y;
    Dimension	width = w->core.width;
    Dimension	height = w->core.height; 

    bzero(w->login.data.name, NAME_LEN);
    bzero(w->login.data.passwd, NAME_LEN);
    bzero(w->login.data.bogus, NAME_LEN);

    XtReleaseGC(gw, w->login.textGC);
    XtReleaseGC(gw, w->login.textBgGC);
    XtReleaseGC(gw, w->login.xorGC);

	if ((w->login.pixmap))
		{ XFreePixmap(dpy, w->login.pixmap); }
}

static void Redisplay(Widget gw, XEvent *event, Region region)
	{ DrawIt((LoginWidget) gw); }

static Boolean SetValues(Widget old, Widget request, Widget self,
    ArgList args, Cardinal *num_args)
{
    LoginWidget	current = (LoginWidget) old;
	LoginWidget	new = (LoginWidget) self;
	Boolean		redisplay = False;
 
	if ((new->core.x != current->core.x))
		{ redisplay = True; }
	if ((new->core.y != current->core.y))
		{ redisplay = True; }
	if ((new->core.width != current->core.width))
		{ redisplay = True; }
	if ((new->core.height != current->core.height))
		{ redisplay = True; }
	if ((new->core.background_pixmap != current->core.background_pixmap))
		{ redisplay = True; }
    return redisplay;
}

char defaultLoginTranslations [] =
"\
Ctrl<Key>H:	delete-previous-character() \n\
Ctrl<Key>D:	delete-character() \n\
Ctrl<Key>B:	move-backward-character() \n\
Ctrl<Key>F:	move-forward-character() \n\
Ctrl<Key>A:	move-to-begining() \n\
Ctrl<Key>E:	move-to-end() \n\
Ctrl<Key>K:	erase-to-end-of-line() \n\
Ctrl<Key>U:	erase-line() \n\
Ctrl<Key>X:	erase-line() \n\
Ctrl<Key>C:	restart-session() \n\
Ctrl<Key>\\\\:	abort-session() \n\
:Ctrl<Key>plus:	allow-all-access() \n\
<Key>BackSpace:	delete-previous-character() \n\
<Key>Delete:	delete-previous-character() \n\
<Key>Return:	finish-field() \n\
<Key>:		insert-char() \
";

XtActionsRec loginActionsTable [] =
{
  { "delete-previous-character",	DeleteBackwardChar},
  { "delete-character",				DeleteForwardChar},
  { "move-backward-character",		MoveBackwardChar},
  { "move-forward-character",		MoveForwardChar},
  { "move-to-begining",				MoveToBegining},
  { "move-to-end",					MoveToEnd},
  { "erase-to-end-of-line",			EraseToEndOfLine},
  { "erase-line",					EraseLine},
  { "finish-field", 				FinishField},
  { "abort-session",				AbortSession},
  { "abort-display",				AbortDisplay},
  { "restart-session",				RestartSession},
  { "insert-char", 					InsertChar},
  { "set-session-argument",			SetSessionArgument},
  { "allow-all-access",				AllowAccess},
};

LoginClassRec loginClassRec =
{
    {
		/* superclass				*/	&widgetClassRec,
		/* class_name				*/	"Login",
		/* size						*/	sizeof(LoginRec),
		/* class_initialize			*/	NULL,
		/* class_part_initialize	*/	NULL,
		/* class_inited				*/	FALSE,
		/* initialize				*/	Initialize,
		/* initialize_hook			*/	NULL,
		/* realize					*/	Realize,
		/* actions					*/	loginActionsTable,
		/* num_actions				*/	XtNumber (loginActionsTable),
		/* resources				*/	resources,
		/* num_resources			*/	XtNumber(resources),
		/* xrm_class				*/	NULLQUARK,
		/* compress_motion			*/	TRUE,
		/* compress_exposure		*/	TRUE,
		/* compress_enterleave		*/	TRUE,
		/* visible_interest			*/	FALSE,
		/* destroy					*/	Destroy,
		/* resize					*/	NULL,
		/* expose					*/	Redisplay,
		/* set_values				*/	SetValues,
		/* set_values_hook			*/	NULL,
		/* set_values_almost		*/	NULL,
		/* get_values_hook			*/	NULL,
		/* accept_focus				*/	NULL,
		/* version					*/	XtVersion,
		/* callback_private			*/	NULL,
		/* tm_table					*/	defaultLoginTranslations,
		/* query_geometry			*/	XtInheritQueryGeometry,
		/* display_accelerator		*/	XtInheritDisplayAccelerator,
		/* extension				*/	NULL
    }
};

WidgetClass loginWidgetClass = (WidgetClass) &loginClassRec;

